﻿Option Explicit On
Option Compare Binary

' RasterPS - Main Routines
' Apr 2013 M.P.

Module ModuleRasterPS

    ' Members:
    ' - Class RPSimgResult : Contains the Result of the Processing Functions
    ' - Class RPSindexedImg: Pixel Access for Indexed Images
    ' - Class RPShistogram : Calculate and display a Histogram and Statistical Data
    ' - Class RPScolorSpace: Calculate Color Space Conversions
    ' - Class RPSpointOps  : Calculate Point Operations
    ' - Class RPSfiltering : Apply Filters and Operators
    ' - Class RPSgeoInterp : Calculate Interpolation for geometric Transformations
    ' - Class Point2Ddbl   : Contains a double 2D-Point
    ' - Class ...

    ' Color-Channels
    Enum enumChannel
        Gray = 0
        Red = 1
        Green = 2
        Blue = 3
        Alpha = 4
        RGB = 5
        Cyan = 11
        Magenta = 12
        Yellow = 13
        CMY = 15
        HSV_Hue = 21
        HSV_Saturation = 22
        HSV_Value = 23
        HSV = 25
        HLS_Hue = 31
        HLS_Luminance = 32
        HLS_Saturation = 33
        HLS = 35
    End Enum

    ' Grayscale Color Weight
    Enum enumGrayColorWeight
        Equal = 1
        Television = 2
        ITU_BT_709 = 3
    End Enum

    ' Weighted Filter
    Enum enumWeightFilter
        RoundFilter = 30
        GaussFilter_3x3 = 31
        LaplaceFilter_3x3 = 32
        GaussFilter_5x5 = 51
        LaplaceFilter_5x5 = 52
    End Enum

    ' Non-Linear Filter
    Enum enumNonLinFilter
        Minimum = -1
        Median = 0
        Maximum = 1
    End Enum

    ' Gradient Filter
    Enum enumGradFilter
        SobelFilter_3x3 = 31
        PrewittFilter_3x3 = 32
    End Enum

    ' Gradient Output
    Enum enumGradOutput
        EdgeStrength = 0
        EdgeOrientation = 1
    End Enum

    ' Morphological Filter
    Enum enumMorphFilter
        Neighbourhood4_3x3 = 20
        Neighbourhood8_3x3 = 30
        DiskFormedFilter_5x5 = 50
    End Enum

    ' Morphological Operator
    Enum enumMorphOper
        Erosion = -1
        Dilatation = 1
    End Enum


    Public Class RPSimgResult
        ' Contains the Result of the Processing Functions
        Public ans As Boolean
        Public img As System.Drawing.Bitmap

    End Class


    Public Class RPSindexedImg
        ' Pixel Access for Indexed Images

        ' Palette Index for Grayscale Images
        Public paletteIndex(255) As Integer

        Public Function GetIndexedPixel(ByVal bmp As Bitmap, ByVal x As Integer, ByVal y As Integer) As Color
            ' Get Pixel
            ' source: http://social.msdn.microsoft.com/Forums/en-US/csharpgeneral/thread/54a096ff-46f3-45ce-8560-bf5a0618ef75/
            ' adapted to Grayscale Images only in Visual Basic
            Dim bmpData As Imaging.BitmapData = bmp.LockBits(New Rectangle(x, y, 1, 1), _
                                         Imaging.ImageLockMode.ReadOnly, _
                                         bmp.PixelFormat)
            Dim c As Color = bmp.Palette.Entries(System.Runtime.InteropServices.Marshal.ReadByte(bmpData.Scan0))
            bmp.UnlockBits(bmpData)

        End Function

        Public Sub SetIndexedPixel(ByVal bmp As Bitmap, ByVal x As Integer, ByVal y As Integer, ByVal c As Color)
            ' Set Pixel
            ' source: http://social.msdn.microsoft.com/Forums/en-US/csharpgeneral/thread/54a096ff-46f3-45ce-8560-bf5a0618ef75/
            ' adapted to Grayscale Images only in Visual Basic
            Dim i As Integer
            Dim min As Integer = Integer.MaxValue
            Dim minI As Integer = 0
            'Dim cPal As Color
            'Dim r As Integer
            'Dim g As Integer
            'Dim b As Integer
            'Dim d As Integer

            '' This Loop makes all too slow !!!
            'For i = 0 To 255
            '    cPal = bmp.Palette.Entries(i)
            '    r = c.R - cPal.R
            '    g = c.G - cPal.G
            '    b = c.B - cPal.B
            '    d = r * r + g * g + b * b
            '    If (d = 0) Then Exit For
            '    If (d < min) Then
            '        min = d
            '        minI = i
            '    End If
            'Next
            'If (i = 256) Then i = minI

            ' from prepared Palette Index for Grayscale Images
            i = paletteIndex(c.R)

            Dim bmpData As Imaging.BitmapData = bmp.LockBits(New Rectangle(x, y, 1, 1), _
                                         Imaging.ImageLockMode.ReadOnly, _
                                         bmp.PixelFormat)
            System.Runtime.InteropServices.Marshal.WriteByte(bmpData.Scan0, CByte(i))
            bmp.UnlockBits(bmpData)

        End Sub

        Public Sub preparePaletteIndex(ByVal bmp As Bitmap)
            ' Prepare Palette Index for Grayscale Images
            Dim i, j As Integer
            Dim min As Integer = Integer.MaxValue
            Dim minI As Integer = 0
            Dim cPal As Color
            Dim r As Integer
            Dim g As Integer
            Dim b As Integer
            Dim d As Integer

            For j = 0 To 255
                For i = 0 To 255
                    cPal = bmp.Palette.Entries(i)
                    r = j - cPal.R
                    g = j - cPal.G
                    b = j - cPal.B
                    d = r * r + g * g + b * b
                    If (d = 0) Then Exit For
                    If (d < min) Then
                        min = d
                        minI = i
                    End If
                Next
                If (minI >= 256) Then minI = 255
                paletteIndex(j) = minI
                min = Integer.MaxValue
                minI = 0
            Next

        End Sub

        Public Function convertFromIndexed(ByVal img As System.Drawing.Bitmap) As System.Drawing.Bitmap
            ' convert from Indexed Image
            ' source: http://www.vbforums.com/showthread.php?368108-Converting-a-256-color-bitmap-to-a-24bit-color-bitmap-in-VB.NET
            Dim w, h As Integer
            Dim con As System.Drawing.Bitmap
            Dim gr As System.Drawing.Graphics

            ' Image-Info
            w = img.Width
            h = img.Height

            ' New Image
            con = New System.Drawing.Bitmap(w, h, System.Drawing.Imaging.PixelFormat.Format24bppRgb)
            gr = System.Drawing.Graphics.FromImage(con)
            gr.DrawImageUnscaled(img, 0, 0)
            gr.Dispose()

            ' Rueckgabe
            convertFromIndexed = con

        End Function

    End Class


    Public Class RPShistogram
        ' Calculate and display a Histogram and Statistical Data
        ' - Function binnedHistogram: Calculate binned Histogram
        ' - Sub plotHistogram: Display Histogram Array as Result
        ' - Sub plotStatistics: Display Statistical Data as Result

        Dim amax As Integer = 255       ' Maximum Color Value
        Public K As Integer = 256       ' number of intensity values RGB and CMY
        Public B As Integer = 256       ' size of histogram, so that K/B is an integer (B=256|128|64|32)
        Public BN As Integer = 250      ' size of histogram for HSV and HLS
        Public HA(B, 4) As Integer      ' histogram array
        Public HE(2, 4) As Integer      ' histogram extrema
        Dim hbins As Integer = 120      ' plot height per channel
        Dim hmarg As Integer = 20       ' plot margin
        Dim mean(4) As Double           ' mean value
        Dim std(4) As Double            ' standard deviation
        Dim Ex(4) As Double             ' Excess
        Dim Sc(4) As Double             ' Scewness
        Dim En(4) As Double             ' Entropy

        Public Function binnedHistogram(ByVal img As System.Drawing.Bitmap, ByVal channel As Integer) As Boolean
            ' Calculate binned Histogram
            ' - img: Input Image
            ' - channel: defined by enumChannel
            Dim locRPScolorSpace As New RPScolorSpace
            Dim w, h, N As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim a(4) As Double
            Dim ai(2) As Integer
            Dim e(2) As Double
            Dim P(4, amax) As Double
            Dim alow(4), ahigh(4) As Integer
            Dim bins(amax) As Integer
            Dim std2(4) As Double
            Dim ad As Double
            Dim i As Integer
            Dim u, v As Integer
            Dim c As Integer

            ' Bild-Info
            w = img.Width
            h = img.Height

            ' Init
            For c = 0 To 4
                ' Extrema
                alow(c) = amax
                ahigh(c) = 0
                ' Statistical Data
                mean(c) = 0
                std2(c) = 0
            Next

            ' Init Bins
            For i = 0 To amax
                bins(i) = i
            Next

            ' Histogramm
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    If (channel = enumChannel.Gray) Then
                        a(0) = (Int(col.R) + Int(col.G) + Int(col.B)) / 3
                    ElseIf (channel = enumChannel.Red) Then
                        a(1) = Int(col.R)
                    ElseIf (channel = enumChannel.Green) Then
                        a(2) = Int(col.G)
                    ElseIf (channel = enumChannel.Blue) Then
                        a(3) = Int(col.B)
                    ElseIf (channel = enumChannel.Alpha) Then
                        a(4) = Int(col.A)
                    ElseIf (channel = enumChannel.RGB) Then
                        a(1) = Int(col.R)
                        a(2) = Int(col.G)
                        a(3) = Int(col.B)
                    ElseIf (channel = enumChannel.CMY) Then
                        ai(0) = Int(col.R)
                        ai(1) = Int(col.G)
                        ai(2) = Int(col.B)
                        e = locRPScolorSpace.RGBtoCMY(ai)
                        a(1) = e(0)
                        a(2) = e(1)
                        a(3) = e(2)
                    ElseIf (channel = enumChannel.HSV) Then
                        ai(0) = Int(col.R)
                        ai(1) = Int(col.G)
                        ai(2) = Int(col.B)
                        e = locRPScolorSpace.RGBtoHSV(ai)
                        a(1) = e(0)
                        a(2) = e(1)
                        a(3) = e(2)
                    ElseIf (channel = enumChannel.HLS) Then
                        ai(0) = Int(col.R)
                        ai(1) = Int(col.G)
                        ai(2) = Int(col.B)
                        e = locRPScolorSpace.RGBtoHLS(ai)
                        a(1) = e(0)
                        a(2) = e(1)
                        a(3) = e(2)
                    End If
                    For c = 0 To 4
                        ' Histogram Bin
                        If (channel <= enumChannel.CMY) Then
                            i = a(c) * B / K
                        Else
                            i = a(c) * BN
                        End If
                        If (i >= 0 And i < B) Then
                            HA(i, c) = HA(i, c) + 1
                        Else
                            MsgBox("Array not appropriate for Color Value", MsgBoxStyle.Exclamation, "Error")
                            binnedHistogram = False
                            Exit Function
                        End If
                        ' Extrema
                        If (a(c) < alow(c)) Then alow(c) = a(c)
                        If (a(c) > ahigh(c)) Then ahigh(c) = a(c)
                        ' Statistical Data
                        mean(c) = mean(c) + a(c)
                        std2(c) = std2(c) + a(c) ^ 2
                    Next
                Next
            Next

            N = w * h   ' Number of Pixels
            For c = 0 To 4
                ' Extrema
                HE(1, c) = alow(c)
                HE(2, c) = ahigh(c)
                ' Statistical Data
                mean(c) = mean(c) / N
                std2(c) = (std2(c) - N * mean(c) ^ 2) / (N - 1)
                std(c) = Math.Sqrt(std2(c))
                ' Scewness, Excess and Entropy
                For i = 0 To amax
                    If (channel <= enumChannel.CMY) Then
                        P(c, i) = bins(i) / N
                    Else
                        P(c, i) = bins(i) / amax / N
                    End If
                    If (P(c, i) = 0) Then
                        P(c, i) = 0.0000000001
                    End If
                Next
            Next

            For c = 0 To 4
                ' Scewness, Excess and Entropy
                Ex(c) = 0
                Sc(c) = 0
                En(c) = 0
                For i = 0 To amax
                    If (channel <= enumChannel.CMY) Then
                        ad = i
                    Else
                        ad = i / amax
                    End If
                    Ex(c) = Ex(c) + (ad - mean(c)) ^ 3 * P(c, i)
                    Sc(c) = Sc(c) + (ad - mean(c)) ^ 4 * P(c, i) - 3
                    If (P(c, i) > 0) Then
                        En(c) = En(c) - P(c, i) * Math.Log(CDbl(P(c, i))) / Math.Log(2.0)
                    End If
                Next
                Ex(c) = Ex(c) / std(c) ^ 3
                Sc(c) = Sc(c) / std2(c) ^ 2
            Next

            ' Rueckgabe
            binnedHistogram = True

        End Function

        Public Sub plotHistogram(ByVal objgr_result As System.Drawing.Graphics, ByVal channel As Integer, ByVal pos As Integer, ByVal vmax As Integer)
            ' Display Histogram Array as Result
            ' - objgr_result: Graphics Object of Drawing Area
            ' - channel: defined by enumChannel
            ' - pos: 1 = top, 2 = center, 3 = bottom
            ' - vmax: vertical Number of Pixels of Drawing Area
            Dim objbr As New System.Drawing.SolidBrush(Color.Black)
            Dim chind As Integer
            Dim title As String
            Dim maxHA As Integer
            Dim uwid As Integer
            Dim uoffset, voffset As Integer
            Dim u1, v1, u2, v2 As Integer
            Dim i As Integer

            ' Check
            If (pos < 1 Or pos > 3) Then
                pos = 1
            End If

            ' Titel
            If (channel = enumChannel.Gray Or channel = enumChannel.Alpha) Then
                title = "Histogram"
            ElseIf ((channel >= enumChannel.Red And channel <= enumChannel.Blue) Or channel = enumChannel.RGB) Then
                title = "Histogram (RGB)"
            ElseIf ((channel >= enumChannel.Cyan And channel <= enumChannel.Yellow) Or channel = enumChannel.CMY) Then
                title = "Histogram (CMY)"
            ElseIf ((channel >= enumChannel.HSV_Hue And channel <= enumChannel.HSV_Value) Or channel = enumChannel.HSV) Then
                title = "Histogram (HSV)"
            ElseIf ((channel >= enumChannel.HLS_Hue And channel <= enumChannel.HLS_Saturation) Or channel = enumChannel.HLS) Then
                title = "Histogram (HLS)"
            Else
                title = "Histogram"
            End If

            ' Channel
            If (channel = enumChannel.Gray) Then
                objbr = New SolidBrush(Color.Black)
                chind = 0
            ElseIf (channel = enumChannel.Red) Then
                objbr = New SolidBrush(Color.Red)
                chind = 1
            ElseIf (channel = enumChannel.Green) Then
                objbr = New SolidBrush(Color.Green)
                chind = 2
            ElseIf (channel = enumChannel.Blue) Then
                objbr = New SolidBrush(Color.Blue)
                chind = 3
            ElseIf (channel = enumChannel.Alpha) Then
                objbr = New SolidBrush(Color.Black)
                chind = 4
            ElseIf (channel = enumChannel.Cyan) Then
                objbr = New SolidBrush(Color.Cyan)
                chind = 1
            ElseIf (channel = enumChannel.Magenta) Then
                objbr = New SolidBrush(Color.Magenta)
                chind = 2
            ElseIf (channel = enumChannel.Yellow) Then
                objbr = New SolidBrush(Color.Yellow)
                chind = 3
            ElseIf (channel = enumChannel.HSV_Hue) Then
                objbr = New SolidBrush(Color.Purple)
                chind = 1
            ElseIf (channel = enumChannel.HSV_Saturation) Then
                objbr = New SolidBrush(Color.Olive)
                chind = 2
            ElseIf (channel = enumChannel.HSV_Value) Then
                objbr = New SolidBrush(Color.Orange)
                chind = 3
            ElseIf (channel = enumChannel.HLS_Hue) Then
                objbr = New SolidBrush(Color.Orchid)
                chind = 1
            ElseIf (channel = enumChannel.HLS_Luminance) Then
                objbr = New SolidBrush(Color.Khaki)
                chind = 2
            ElseIf (channel = enumChannel.HLS_Saturation) Then
                objbr = New SolidBrush(Color.Teal)
                chind = 3
            End If

            ' Position
            uoffset = 1
            voffset = pos * (hbins + hmarg)

            ' Skalierung
            maxHA = 0
            For i = 0 To B - 1
                If (HA(i, chind) > maxHA) Then
                    maxHA = HA(i, chind)
                End If
            Next

            ' Check
            If (maxHA = 0) Then
                Exit Sub

            End If

            ' Plot Histogram
            For i = 0 To B - 1
                u1 = K / B * i
                v1 = 0
                u2 = K / B * i
                v2 = HA(i, chind) / maxHA * hbins
                uwid = Int(K / B)
                objgr_result.FillRectangle(objbr, uoffset + u1, voffset - v2, uwid, v2)
            Next
            objgr_result.DrawLine(New Pen(Color.Black), uoffset, voffset, uoffset + uwid * B, voffset)
            objgr_result.DrawString(title, New Font(FontFamily.GenericSerif, 10), Brushes.Black, 0, 0)

        End Sub

        Public Sub plotStatistics(ByVal objgr_result As System.Drawing.Graphics, ByVal channel As Integer, ByVal pos As Integer, ByVal vmax As Integer)
            ' Display Statistical Data as Result
            ' - objgr_result: Graphics Object of Drawing Area
            ' - channel: defined by enumChannel
            ' - pos: 1 = top, 2 = center, 3 = bottom
            ' - vmax: vertical Number of Pixels of Drawing Area
            Dim objbr As New System.Drawing.SolidBrush(Color.Black)
            Dim chind As Integer
            Dim title As String
            Dim vstep As Integer
            Dim uoffset, voffset As Integer

            ' Check
            If (pos < 1 Or pos > 3) Then
                pos = 1
            End If

            ' Titel
            If (channel = enumChannel.Gray Or channel = enumChannel.Alpha) Then
                title = "Statistical Data"
            ElseIf ((channel >= enumChannel.Red And channel <= enumChannel.Blue) Or channel = enumChannel.RGB) Then
                title = "Statistical Data (RGB)"
            ElseIf ((channel >= enumChannel.Cyan And channel <= enumChannel.Yellow) Or channel = enumChannel.CMY) Then
                title = "Statistical Data (CMY)"
            ElseIf ((channel >= enumChannel.HSV_Hue And channel <= enumChannel.HSV_Value) Or channel = enumChannel.HSV) Then
                title = "Statistical Data (HSV)"
            ElseIf ((channel >= enumChannel.HLS_Hue And channel <= enumChannel.HLS_Saturation) Or channel = enumChannel.HLS) Then
                title = "Statistical Data (HLS)"
            Else
                title = "Histogram"
            End If

            ' Channel
            If (channel = enumChannel.Gray) Then
                objbr = New SolidBrush(Color.Black)
                chind = 0
            ElseIf (channel = enumChannel.Red) Then
                objbr = New SolidBrush(Color.Red)
                chind = 1
            ElseIf (channel = enumChannel.Green) Then
                objbr = New SolidBrush(Color.Green)
                chind = 2
            ElseIf (channel = enumChannel.Blue) Then
                objbr = New SolidBrush(Color.Blue)
                chind = 3
            ElseIf (channel = enumChannel.Alpha) Then
                objbr = New SolidBrush(Color.Black)
                chind = 4
            ElseIf (channel = enumChannel.Cyan) Then
                objbr = New SolidBrush(Color.Cyan)
                chind = 1
            ElseIf (channel = enumChannel.Magenta) Then
                objbr = New SolidBrush(Color.Magenta)
                chind = 2
            ElseIf (channel = enumChannel.Yellow) Then
                objbr = New SolidBrush(Color.Yellow)
                chind = 3
            ElseIf (channel = enumChannel.HSV_Hue) Then
                objbr = New SolidBrush(Color.Purple)
                chind = 1
            ElseIf (channel = enumChannel.HSV_Saturation) Then
                objbr = New SolidBrush(Color.Olive)
                chind = 2
            ElseIf (channel = enumChannel.HSV_Value) Then
                objbr = New SolidBrush(Color.Orange)
                chind = 3
            ElseIf (channel = enumChannel.HLS_Hue) Then
                objbr = New SolidBrush(Color.Orchid)
                chind = 1
            ElseIf (channel = enumChannel.HLS_Luminance) Then
                objbr = New SolidBrush(Color.Khaki)
                chind = 2
            ElseIf (channel = enumChannel.HLS_Saturation) Then
                objbr = New SolidBrush(Color.Teal)
                chind = 3
            End If

            ' Position
            vstep = 20
            uoffset = 5
            voffset = pos * (hbins + hmarg)

            ' Plot Data
            objgr_result.DrawString("Mean Val.: " & Format(mean(chind), "0.00"), New Font(FontFamily.GenericSerif, 10), objbr, uoffset, voffset - 5 * vstep)
            objgr_result.DrawString("Std Dev.  : " & Format(std(chind), "0.00"), New Font(FontFamily.GenericSerif, 10), objbr, uoffset, voffset - 4 * vstep)
            objgr_result.DrawString("Scewness: " & Format(Sc(chind), "0.00"), New Font(FontFamily.GenericSerif, 10), objbr, uoffset, voffset - 3 * vstep)
            objgr_result.DrawString("Excess      : " & Format(Ex(chind), "0.00"), New Font(FontFamily.GenericSerif, 10), objbr, uoffset, voffset - 2 * vstep)
            objgr_result.DrawString("Entropy    : " & Format(En(chind), "0.00"), New Font(FontFamily.GenericSerif, 10), objbr, uoffset, voffset - 1 * vstep)
            objgr_result.DrawString(title, New Font(FontFamily.GenericSerif, 10), Brushes.Black, 0, 0)

        End Sub

        Public Sub plotImageInfo(ByVal objgr_result As System.Drawing.Graphics, ByVal img As System.Drawing.Bitmap)
            ' Display Image Info
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer
            Dim imgFormat, horResol As String
            Dim vstep As Integer
            Dim uoffset, voffset As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height
            imgFormat = img.PixelFormat.ToString
            horResol = img.HorizontalResolution.ToString

            ' Position
            vstep = 20
            uoffset = 5
            voffset = 4 * (hbins + hmarg)

            objgr_result.DrawString("Image Info", New Font(FontFamily.GenericSerif, 10), Brushes.Black, uoffset, voffset - 5 * vstep)
            objgr_result.DrawString("- Width : " & w.ToString, New Font(FontFamily.GenericSerif, 10), Brushes.Black, uoffset, voffset - 4 * vstep)
            objgr_result.DrawString("- Height: " & h.ToString, New Font(FontFamily.GenericSerif, 10), Brushes.Black, uoffset, voffset - 3 * vstep)
            objgr_result.DrawString("- Format: " & imgFormat, New Font(FontFamily.GenericSerif, 10), Brushes.Black, uoffset, voffset - 2 * vstep)
            objgr_result.DrawString("- Resol. : " & horResol, New Font(FontFamily.GenericSerif, 10), Brushes.Black, uoffset, voffset - 1 * vstep)

        End Sub

    End Class


    Public Class RPScolorSpace
        ' Calculate Color Space Conversions
        ' - Function RGBtoHSV: Conversion from RGB to HSV
        ' - Function HSVtoRGB: Conversion from HSV to RGB
        ' - Function RGBtoHLS: Conversion from RGB to HLS
        ' - Function HLStoRGB: Conversion from HLS to RGB
        ' - Function RGBtoYUV: Conversion from RGB to YUV
        ' - Function YUVtoRGB: Conversion from YUV to RGB
        ' - Function RGBtoYCC: Conversion from RGB to YCC
        ' - Function YCCtoRGB: Conversion from YCC to RGB
        ' - Function RGBtoCMY: Conversion from RGB to CMY
        ' - Function CMYtoRGB: Conversion from CMY to RGB

        Dim amax As Integer = 255       ' Maximum Color Value

        Public Function RGBtoHSV(ByVal a() As Integer) As Double()
            ' Conversion from RGB to HSV
            ' - a(): RGB Values at indices (0,1,2)
            Dim e(2) As Double
            Dim R, G, B As Integer
            Dim H, S, V As Double
            Dim rr, gg, bb, hh As Double
            Dim cMax As Double = 255.0
            Dim cHi, cLo, cRng As Integer

            ' Eingabe
            R = a(0) : G = a(1) : B = a(2)

            ' highest and lowest Color Value, Color Range
            cHi = Math.Max(R, Math.Max(G, B))
            cLo = Math.Min(R, Math.Min(G, B))
            cRng = cHi - cLo

            ' Value V
            V = cHi / cMax

            ' Saturation S
            If (cHi > 0) Then
                S = CDbl(cRng / cHi)
            Else
                S = 0
            End If

            ' Hue H
            If (cRng > 0) Then
                rr = CDbl((cHi - R) / cRng)
                gg = CDbl((cHi - G) / cRng)
                bb = CDbl((cHi - B) / cRng)
                If (R = cHi) Then
                    hh = bb - gg
                ElseIf (G = cHi) Then
                    hh = rr - bb + 2.0
                ElseIf (B = cHi) Then
                    hh = gg - rr + 4.0
                End If
                If (hh < 0) Then
                    hh = hh + 6
                End If
                H = hh / 6
            Else
                H = 0
            End If

            ' Rueckgabe
            e(0) = H
            e(1) = S
            e(2) = V
            RGBtoHSV = e

        End Function

        Public Function HSVtoRGB(ByVal e() As Double) As Integer()
            ' Conversion from HSV to RGB
            ' - e(): HSV Values at indices (0,1,2)
            Dim a(2) As Integer
            Dim H, S, V As Double
            Dim R, G, B As Integer
            Dim rr, gg, bb, hh As Double
            Dim N As Integer = 256
            Dim c1 As Integer
            Dim c2 As Double
            Dim x, y, z As Double

            ' Eingabe
            H = e(0) : S = e(1) : V = e(2)

            ' intermediate Values
            hh = (6 * H) Mod 6
            c1 = Fix(hh)    ' Fix !!!
            c2 = hh - c1
            x = (1 - S) * V
            y = (1 - (S * c2)) * V
            z = (1 - (S * (1 - c2))) * V

            ' normalized RGB Values
            Select Case c1
                Case 0
                    rr = V
                    gg = z
                    bb = x
                    Exit Select
                Case 1
                    rr = y
                    gg = V
                    bb = x
                    Exit Select
                Case 2
                    rr = x
                    gg = V
                    bb = z
                    Exit Select
                Case 3
                    rr = x
                    gg = y
                    bb = V
                    Exit Select
                Case 4
                    rr = z
                    gg = x
                    bb = V
                    Exit Select
                Case 5
                    rr = V
                    gg = x
                    bb = y
                    Exit Select
            End Select

            ' RGB Values
            R = Math.Min(Math.Round(rr * N), N - 1)
            G = Math.Min(Math.Round(gg * N), N - 1)
            B = Math.Min(Math.Round(bb * N), N - 1)

            ' Rueckgabe
            a(0) = R
            a(1) = G
            a(2) = B
            HSVtoRGB = a

        End Function

        Public Function RGBtoHLS(ByVal a() As Integer) As Double()
            ' Conversion from RGB to HLS
            ' - a(): RGB Values at indices (0,1,2)
            Dim e(2) As Double
            Dim R, G, B As Double   ' RGB c [0,1] !!!
            Dim H, L, S As Double
            Dim rr, gg, bb, hh As Double
            Dim cMax As Double = 255.0
            Dim d As Double
            Dim cHi, cLo, cRng As Double

            ' Eingabe
            R = a(0) / amax : G = a(1) / amax : B = a(2) / amax

            ' highest and lowest Color Value, Color Range
            cHi = Math.Max(R, Math.Max(G, B))
            cLo = Math.Min(R, Math.Min(G, B))
            cRng = cHi - cLo

            ' Luminance L
            L = (cHi + cLo) / 2

            ' Saturation S
            If (L > 0 And L < 1) Then
                If (L <= 0.5) Then
                    d = L
                Else
                    d = (1 - L)
                End If
                S = 0.5 * cRng / d
            Else
                S = 0
            End If

            ' Hue H
            If (cHi > 0 And cRng > 0) Then
                rr = CDbl((cHi - R) / cRng)
                gg = CDbl((cHi - G) / cRng)
                bb = CDbl((cHi - B) / cRng)
                If (R = cHi) Then
                    hh = bb - gg
                ElseIf (G = cHi) Then
                    hh = rr - bb + 2.0
                ElseIf (B = cHi) Then
                    hh = gg - rr + 4.0
                End If
                If (hh < 0) Then
                    hh = hh + 6
                End If
                H = hh / 6
            Else
                H = 0
            End If

            ' Rueckgabe
            e(0) = H
            e(1) = L
            e(2) = S
            RGBtoHLS = e

        End Function

        Public Function HLStoRGB(ByVal e() As Double) As Integer()
            ' Conversion from HLS to RGB
            ' - e(): HLS Values at indices (0,1,2)
            Dim a(2) As Integer
            Dim H, L, S As Double
            Dim R, G, B As Double   ' RGB c [0,1] !!!
            Dim hh As Double
            Dim c1 As Integer
            Dim c2, d As Double
            Dim w, x, y, z As Double

            ' Eingabe
            H = e(0) : L = e(1) : S = e(2)

            If (L <= 0) Then
                ' Black
                R = 0
                G = 0
                B = 0
            ElseIf (L >= 1) Then
                ' White
                R = 1
                G = 1
                B = 1
            Else
                ' Color
                hh = (6 * H) Mod 6
                c1 = Fix(hh)   ' Fix !!!
                c2 = hh - c1
                If (L <= 0.5) Then
                    d = (S * L)
                Else
                    d = (S * (1 - L))
                End If
                W = L + d
                x = L - d
                y = w - (w - x) * c2
                z = x + (w - x) * c2
                ' RGB Values
                Select Case c1
                    Case 0
                        R = w
                        G = z
                        B = x
                        Exit Select
                    Case 1
                        R = y
                        G = w
                        B = x
                        Exit Select
                    Case 2
                        R = x
                        G = w
                        B = z
                        Exit Select
                    Case 3
                        R = x
                        G = y
                        B = w
                        Exit Select
                    Case 4
                        R = z
                        G = x
                        B = w
                        Exit Select
                    Case 5
                        R = w
                        G = x
                        B = y
                        Exit Select
                End Select
            End If

            ' Rueckgabe
            a(0) = R * amax
            a(1) = G * amax
            a(2) = B * amax
            HLStoRGB = a

        End Function

        Public Function RGBtoYUV(ByVal a() As Integer) As Double()
            ' Conversion from RGB to YUV
            ' - a(): RGB Values at indices (0,1,2)
            Dim e(2) As Double
            Dim R, G, B As Integer
            Dim Y, U, V As Double
            Dim CM(,) As Double = {{0.299, 0.587, 0.114}, {-0.147, -0.289, 0.436}, {0.615, -0.515, -0.1}}
            Dim c As Integer

            ' Eingabe
            R = a(0) : G = a(1) : B = a(2)

            ' Apply Transformation Matrix
            Y = 0
            For c = 0 To 2
                Y = Y + CM(0, c) * a(c)
            Next
            U = 0
            For c = 0 To 2
                U = U + CM(1, c) * a(c)
            Next
            V = 0
            For c = 0 To 2
                V = V + CM(2, c) * a(c)
            Next

            ' Rueckgabe
            e(0) = Y
            e(1) = U
            e(2) = V
            RGBtoYUV = e

        End Function

        Public Function YUVtoRGB(ByVal e() As Double) As Integer()
            ' Conversion from YUV to RGB
            ' - e(): YUV Values at indices (0,1,2)
            Dim a(2) As Integer
            Dim Y, U, V As Double
            Dim R, G, B As Double   ' sum !!!
            Dim CM(,) As Double = {{1.0, 0.0, 1.14}, {1.0, -0.395, -0.581}, {1.0, 2.032, 0.0}}
            Dim c As Integer

            ' Eingabe
            Y = e(0) : U = e(1) : V = e(2)

            ' Apply Transformation Matrix
            R = 0
            For c = 0 To 2
                R = R + CM(0, c) * e(c)
            Next
            G = 0
            For c = 0 To 2
                G = G + CM(1, c) * e(c)
            Next
            B = 0
            For c = 0 To 2
                B = B + CM(2, c) * e(c)
            Next

            ' Rueckgabe
            a(0) = CInt(R)
            a(1) = CInt(G)
            a(2) = CInt(B)
            YUVtoRGB = a

        End Function

        Public Function RGBtoYCC(ByVal a() As Integer) As Double()
            ' Conversion from RGB to YCC
            ' - a(): RGB Values at indices (0,1,2)
            Dim e(2) As Double
            Dim R, G, B As Integer
            Dim Y, Cr, Cb As Double
            Dim CM(,) As Double = {{0.299, 0.587, 0.114}, {-0.169, -0.331, 0.5}, {0.5, -0.419, -0.081}}
            Dim c As Integer

            ' Eingabe
            R = a(0) : G = a(1) : B = a(2)

            ' Apply Transformation Matrix
            Y = 0
            For c = 0 To 2
                Y = Y + CM(0, c) * a(c)
            Next
            Cr = 0
            For c = 0 To 2
                Cr = Cr + CM(1, c) * a(c)
            Next
            Cb = 0
            For c = 0 To 2
                Cb = Cb + CM(2, c) * a(c)
            Next

            ' Rueckgabe
            e(0) = Y
            e(1) = Cr
            e(2) = Cb
            RGBtoYCC = e

        End Function

        Public Function YCCtoRGB(ByVal e() As Double) As Integer()
            ' Conversion from YCC to RGB
            ' - e(): YCC Values at indices (0,1,2)
            Dim a(2) As Integer
            Dim Y, Cr, Cb As Double
            Dim R, G, B As Double   ' sum !!!
            Dim CM(,) As Double = {{1.0, 0.0, 1.403}, {1.0, -0.344, -0.714}, {1.0, 1.773, 0.0}}
            Dim c As Integer

            ' Eingabe
            Y = e(0) : Cr = e(1) : Cb = e(2)

            ' Apply Transformation Matrix
            R = 0
            For c = 0 To 2
                R = R + CM(0, c) * e(c)
            Next
            G = 0
            For c = 0 To 2
                G = G + CM(1, c) * e(c)
            Next
            B = 0
            For c = 0 To 2
                B = B + CM(2, c) * e(c)
            Next

            ' Rueckgabe
            a(0) = CInt(R)
            a(1) = CInt(G)
            a(2) = CInt(B)
            YCCtoRGB = a

        End Function

        Public Function RGBtoCMY(ByVal a() As Integer) As Double()
            ' Conversion from RGB to CMY
            ' - a(): RGB Values at indices (0,1,2)
            Dim e(2) As Double
            Dim R, G, B As Integer
            Dim C, M, Y As Double

            ' Eingabe
            R = a(0) : G = a(1) : B = a(2)

            ' CMY Values
            C = amax - R
            M = amax - G
            Y = amax - B

            ' Rueckgabe
            e(0) = C
            e(1) = M
            e(2) = Y
            RGBtoCMY = e

        End Function

        Public Function CMYtoRGB(ByVal e() As Double) As Integer()
            ' Conversion from CMY to RGB
            ' - e(): CMY Values at indices (0,1,2)
            Dim a(2) As Integer
            Dim C, M, Y As Double
            Dim R, G, B As Integer

            ' Eingabe
            C = e(0) : M = e(1) : Y = e(2)

            ' RGB Values
            R = amax - C
            G = amax - M
            B = amax - Y

            ' Rueckgabe
            a(0) = R
            a(1) = G
            a(2) = B
            CMYtoRGB = a

        End Function

    End Class


    Public Class RPSpointOps
        ' Calculate Point Operations
        ' - Function invertImage: Calculate Inverse Image
        ' - Function binaryImage: Calculate Binary Image
        ' - Function grayscaleImage: Calculate Grayscale Image
        ' - Function changeBrightness: Change Brightness of Image
        ' - Function autoContrast: Calculate Auto Contrast
        ' - Function histoEqualization: Calculate Histogram Equalization
        ' - Function gammaCorrection: Calculate Gamma Correction
        ' - Function imageQuantization: Calculate Image Quantization
        ' - Function imageDesaturation: Calculate Image Desaturation

        Dim amax As Integer = 255       ' Maximum Color Value
        Dim K As Integer = 256          ' number of intensity values
        Public imgbak As System.Drawing.Bitmap

        Public Function invertImage(ByVal img As System.Drawing.Bitmap) As RPSimgResult
            ' Calculate Inverse Image
            ' - img: Input Image
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colinv As System.Drawing.Color
            Dim a(3) As Integer
            Dim u, v As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                invertImage = locRPSimgResult
                Exit Function

            End If

            ' Image Inversion
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    a(1) = amax - Int(col.R)
                    a(2) = amax - Int(col.G)
                    a(3) = amax - Int(col.B)
                    colinv = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colinv)
                    Else
                        img.SetPixel(u, v, colinv)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            invertImage = locRPSimgResult

        End Function

        Public Function grayscaleImage(ByVal img As System.Drawing.Bitmap, ByVal wtype As Integer) As RPSimgResult
            ' Calculate Grayscale Image
            ' - img: Input Image
            ' - wtype: Weighting Type: 1=Equal, 2=Television, 3=ITU-BT.709
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colinv As System.Drawing.Color
            Dim a(3) As Integer
            Dim cw(3) As Double
            Dim mean As Integer
            Dim u, v As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                grayscaleImage = locRPSimgResult
                Exit Function

            End If

            ' Color Weights (sum = 1.0)
            If (wtype = 1) Then
                ' Equal
                cw(1) = 1
                cw(2) = 1
                cw(3) = 1
            ElseIf (wtype = 2) Then
                ' Television
                cw(1) = 0.299
                cw(2) = 0.587
                cw(3) = 0.114
            ElseIf (wtype = 3) Then
                ' ITU-BT.709
                cw(1) = 0.2125
                cw(2) = 0.7154
                cw(3) = 0.0721
            Else
                ' Equal
                cw(1) = 1
                cw(2) = 1
                cw(3) = 1
            End If

            ' Grayscale Image
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    mean = Int((cw(1) * Int(col.R) + cw(2) * Int(col.G) + cw(3) * Int(col.B)) / 3)
                    a(1) = mean
                    a(2) = mean
                    a(3) = mean
                    colinv = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colinv)
                    Else
                        img.SetPixel(u, v, colinv)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            grayscaleImage = locRPSimgResult

        End Function

        Public Function binaryImage(ByVal img As System.Drawing.Bitmap) As RPSimgResult
            ' Calculate Binary Image
            ' - img: Input Image
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colinv As System.Drawing.Color
            Dim a(3) As Integer
            Dim mean As Integer
            Dim u, v As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                binaryImage = locRPSimgResult
                Exit Function

            End If

            ' Binary Image
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    mean = Int((Int(col.R) + Int(col.G) + Int(col.B)) / 3)
                    If (mean < 128) Then
                        a(1) = 0
                        a(2) = 0
                        a(3) = 0
                    Else
                        a(1) = amax
                        a(2) = amax
                        a(3) = amax
                    End If
                    colinv = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colinv)
                    Else
                        img.SetPixel(u, v, colinv)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            binaryImage = locRPSimgResult

        End Function

        Public Function changeBrightness(ByVal img As System.Drawing.Bitmap, ByVal sgn As Integer, ByVal ashft As Integer) As RPSimgResult
            ' Change Brightness of Image
            ' - img: Input Image
            ' - sgn > 0: encrease Brightness, sgn < 0: decrease Brightness
            ' - ashft  : Value of Change
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colchg As System.Drawing.Color
            Dim a(3) As Integer
            Dim u, v As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                changeBrightness = locRPSimgResult
                Exit Function

            End If

            ' Check
            If (sgn <> -1 And sgn <> 1) Then sgn = 1
            If (ashft > amax) Then ashft = amax
            If (ashft < 0) Then ashft = 0

            ' Brightness Shift
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    a(1) = Int(col.R) + sgn * ashft
                    a(2) = Int(col.G) + sgn * ashft
                    a(3) = Int(col.B) + sgn * ashft
                    For c = 1 To 3
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colchg = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colchg)
                    Else
                        img.SetPixel(u, v, colchg)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            changeBrightness = locRPSimgResult

        End Function

        Public Function autoContrast(ByVal img As System.Drawing.Bitmap) As RPSimgResult
            ' Calculate Auto Contrast
            ' - img: Input Image
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colcnt As System.Drawing.Color
            Dim a(3) As Integer
            Dim alow(4), ahigh(4) As Integer
            Dim u, v As Integer
            Dim histo As New RPShistogram
            Dim ret As Boolean

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                autoContrast = locRPSimgResult
                Exit Function

            End If

            ' Extrema from Histogram
            ret = histo.binnedHistogram(img, enumChannel.RGB)
            If (ret = False) Then
                MsgBox("Not possible for this Image", MsgBoxStyle.Information, "Histogram Error")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                autoContrast = locRPSimgResult
                Exit Function

            End If
            For c = 1 To 3
                alow(c) = histo.HE(1, c)
                ahigh(c) = histo.HE(2, c)
            Next

            ' Auto Contrast
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    a(1) = (Int(col.R) - alow(1)) * amax / (ahigh(1) - alow(1))
                    a(2) = (Int(col.G) - alow(2)) * amax / (ahigh(2) - alow(2))
                    a(3) = (Int(col.B) - alow(3)) * amax / (ahigh(3) - alow(3))
                    For c = 1 To 3
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colcnt = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colcnt)
                    Else
                        img.SetPixel(u, v, colcnt)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            autoContrast = locRPSimgResult

        End Function

        Public Function histoEqualization(ByVal img As System.Drawing.Bitmap) As RPSimgResult
            ' Calculate Histogram Equalization
            ' - img: Input Image
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h, N As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colequ As System.Drawing.Color
            Dim a(3) As Integer
            Dim HC(,) As Integer
            Dim u, v As Integer
            Dim i As Integer
            Dim c As Integer
            Dim histo As New RPShistogram
            Dim ret As Boolean

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                histoEqualization = locRPSimgResult
                Exit Function

            End If

            ' Histogram
            ret = histo.binnedHistogram(img, enumChannel.RGB)
            If (ret = False) Then
                MsgBox("Not possible for this Image", MsgBoxStyle.Information, "Histogram Error")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                histoEqualization = locRPSimgResult
                Exit Function

            End If

            ' Cumulative Histogram 
            ReDim HC(histo.B, 3)
            HC(0, 1) = histo.HA(0, 1)
            HC(0, 2) = histo.HA(0, 2)
            HC(0, 3) = histo.HA(0, 3)
            For i = 1 To histo.B - 1
                For c = 1 To 3
                    HC(i, c) = histo.HA(i, c)
                    HC(i, c) = HC(i - 1, c) + HC(i, c)
                Next
            Next

            ' Histogram Equalization
            N = w * h   ' Number of Pixels
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    i = col.R * histo.B / histo.K
                    a(1) = HC(i, 1) * (histo.K - 1) / N
                    i = col.G * histo.B / histo.K
                    a(2) = HC(i, 2) * (histo.K - 1) / N
                    i = col.B * histo.B / histo.K
                    a(3) = HC(i, 3) * (histo.K - 1) / N
                    colequ = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colequ)
                    Else
                        img.SetPixel(u, v, colequ)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            histoEqualization = locRPSimgResult

        End Function

        Public Function gammaCorrection(ByVal img As System.Drawing.Bitmap, ByVal gamma As Double) As RPSimgResult
            ' Calculate Gamma Correction
            ' - img: Input Image
            ' - gamma: Gamma Value (0.25 ... 4.0)
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colgam As System.Drawing.Color
            Dim a(3) As Integer
            Dim LUT(amax) As Integer
            Dim aa, bb As Double
            Dim u, v As Integer
            Dim l As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                gammaCorrection = locRPSimgResult
                Exit Function

            End If

            ' Check
            If (gamma < 0.0) Then gamma = 0.0

            ' Lookup Table
            For l = 0 To amax - 1
                aa = l / amax                           ' scale to [0,1]
                bb = Math.Pow(aa, gamma)                ' gamma function
                LUT(l) = CInt(Math.Round(bb * amax))    ' scale to [0,255]
            Next

            ' Gamma Correction
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    a(1) = LUT(col.R)
                    a(2) = LUT(col.G)
                    a(3) = LUT(col.B)
                    colgam = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colgam)
                    Else
                        img.SetPixel(u, v, colgam)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            gammaCorrection = locRPSimgResult

        End Function

        Public Function imageQuantization(ByVal img As System.Drawing.Bitmap, ByVal numSteps As Integer) As RPSimgResult
            ' Calculate Image Quantization
            ' - img: Input Image
            ' - numSteps: Number of Steps (1 ... 256)
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colqun As System.Drawing.Color
            Dim a(3) As Integer
            Dim LUT(amax) As Integer
            Dim stwith As Integer
            Dim alower As Integer
            Dim aupper As Integer
            Dim steps As Integer
            Dim amean As Integer
            Dim u, v As Integer
            Dim l As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                imageQuantization = locRPSimgResult
                Exit Function

            End If

            ' Check
            If (numSteps < 1) Then numSteps = 1
            If (numSteps > 256) Then numSteps = 256

            ' Init
            stwith = Fix(K / (numSteps))
            alower = 0
            aupper = Fix(stwith / 2)
            steps = 1
            amean = 0       ' die erste Stufe muss schwarz werden

            ' Lookup Table
            For l = 0 To amax
                LUT(l) = 0
            Next

            ' Iteration
            While (steps <= numSteps)
                For l = alower To aupper - 1
                    LUT(l) = amean
                Next
                alower = aupper                 ' "Anfang" und ...
                aupper = aupper + stwith        ' "Ende" der aktuellen Stufe
                amean = amean + stwith          ' neuer mittlerer Grauwert
                If (amean > amax) Then          ' Sicherheit clamping
                    amean = amax
                End If
                steps = steps + 1
            End While

            ' die letzte Stufe auf weiss setzen
            For l = alower To amax
                LUT(l) = amax
            Next

            ' Image Quantization
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    a(1) = LUT(col.R)
                    a(2) = LUT(col.G)
                    a(3) = LUT(col.B)
                    colqun = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colqun)
                    Else
                        img.SetPixel(u, v, colqun)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            imageQuantization = locRPSimgResult

        End Function

        Public Function imageDesaturation(ByVal img As System.Drawing.Bitmap, ByVal dfac As Double) As RPSimgResult
            ' Calculate Image Desaturation
            ' - img: Input Image
            ' - dfac: Desaturation Factor (0.0 ... 1.0)
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colinv As System.Drawing.Color
            Dim a(3) As Integer
            Dim mean As Integer
            Dim u, v As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                imageDesaturation = locRPSimgResult
                Exit Function
            End If

            ' Check
            If (dfac < 0.0) Then dfac = 0.0
            If (dfac > 1.0) Then dfac = 1.0

            ' Image Desaturation
            For v = 0 To h - 1
                For u = 0 To w - 1
                    col = img.GetPixel(u, v)
                    mean = Int((Int(col.R) + Int(col.G) + Int(col.B)) / 3)
                    a(1) = mean + dfac * (Int(col.R) - mean)
                    a(2) = mean + dfac * (Int(col.G) - mean)
                    a(3) = mean + dfac * (Int(col.B) - mean)
                    For c = 1 To 3
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colinv = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colinv)
                    Else
                        img.SetPixel(u, v, colinv)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            imageDesaturation = locRPSimgResult

        End Function

    End Class


    Public Class RPSfiltering
        ' Filtering
        ' - Function boxFilter: Apply Box Filter
        ' - Function weightedFilter: Apply Weighted Filter
        ' - Function minmaxMedianFilter: Apply Minimum/Maximum/Median Filter
        ' - Function unsharpMasking: Unsharp Masking from original and filtered Image
        ' - Function edgeSharpening: Edge Sharpening from original and filtered Image
        ' - Function gradientFilter: Apply Gradient Filter
        ' - Function morphologFilter: Apply Morphological Filter

        Dim amax As Integer = 255       ' Maximum Color Value
        Dim K As Integer = 256          ' number of intensity values
        Public imgbak As System.Drawing.Bitmap

        Public Function boxFilter(ByVal img As System.Drawing.Bitmap) As RPSimgResult
            ' Apply Box Filter
            ' - img: Input Image
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim wrk As System.Drawing.Bitmap
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim sum(3) As Integer
            Dim u, v As Integer
            Dim i, j As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                boxFilter = locRPSimgResult
                Exit Function

            End If

            ' Work Copy
            wrk = img.Clone

            ' Create Filter
            Dim filter(,) As Double = {{1, 1, 1}, {1, 1, 1}, {1, 1, 1}}

            ' Apply Box Filter
            For v = 1 To h - 2
                For u = 1 To w - 2
                    sum(1) = 0
                    sum(2) = 0
                    sum(3) = 0
                    For i = -1 To 1
                        For j = -1 To 1
                            col = wrk.GetPixel(u + i, v + j)
                            sum(1) = sum(1) + Int(col.R)
                            sum(2) = sum(2) + Int(col.G)
                            sum(3) = sum(3) + Int(col.B)
                        Next
                    Next
                    For c = 1 To 3
                        a(c) = Int(Math.Round(sum(c) / 9))
                    Next
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colfil)
                    Else
                        img.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            boxFilter = locRPSimgResult

        End Function

        Public Function weightedFilter(ByVal img As System.Drawing.Bitmap, ByVal filnum As Integer) As RPSimgResult
            ' Apply Weighted Filter
            ' - img: Input Image
            ' - filnum: 30="Round" Filter, 31=Gauss Filter 3x3, 32=Laplace Filter 3x3, 
            '           51=Gauss Filter 5x5, 52=Laplace Filter 5x5, else Simple Filter
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim wrk As System.Drawing.Bitmap
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim sum(3) As Integer
            Dim u, v As Integer
            Dim i, j As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                weightedFilter = locRPSimgResult
                Exit Function

            End If

            ' Work Copy
            wrk = img.Clone

            ' Create Filter
            Dim Filterx(,) As Double
            Dim FilterSum As Integer
            Dim FilterSize As Integer
            If (filnum = 30) Then
                ' "Round" Filter
                FilterSize = 2
                FilterSum = 1
                Dim Filtery(,) As Double = {{0.075, 0.125, 0.075}, {0.125, 0.2, 0.125}, {0.075, 0.125, 0.075}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 31) Then
                ' Gauss Filter 3x3
                FilterSize = 2
                FilterSum = 6
                Dim Filtery(,) As Double = {{0, 1, 0}, {1, 2, 1}, {0, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 32) Then
                ' Laplace Filter 3x3
                FilterSize = 2
                FilterSum = 0
                Dim Filtery(,) As Double = {{0, 1, 0}, {1, -4, 1}, {0, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 51) Then
                ' Gauss Filter 5x5
                FilterSize = 4
                FilterSum = 57
                Dim Filtery(,) As Double = {{0, 1, 2, 1, 0}, {1, 3, 5, 3, 1}, {2, 5, 9, 5, 2}, {1, 3, 5, 3, 1}, {0, 1, 2, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 52) Then
                ' Laplace Filter 5x5
                FilterSize = 4
                FilterSum = 0
                Dim Filtery(,) As Double = {{0, 0, -1, 0, 0}, {0, -1, -2, -1, 0}, {-1, -2, 16, -2, -1}, {0, -1, -2, -1, 0}, {0, 0, -1, 0, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            Else
                ' Simple Filter
                FilterSize = 2
                FilterSum = 1
                Dim Filtery(,) As Double = {{0.1, 0.1, 0.1}, {0.1, 0.2, 0.1}, {0.1, 0.1, 0.1}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            End If

            ' Apply Weighted Filter
            For v = FilterSize / 2 To h - 1 - FilterSize / 2
                For u = FilterSize / 2 To w - 1 - FilterSize / 2
                    sum(1) = 0
                    sum(2) = 0
                    sum(3) = 0
                    For i = -FilterSize / 2 To FilterSize / 2
                        For j = -FilterSize / 2 To FilterSize / 2
                            col = wrk.GetPixel(u + i, v + j)
                            sum(1) = sum(1) + Int(col.R * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                            sum(2) = sum(2) + Int(col.G * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                            sum(3) = sum(3) + Int(col.B * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                        Next
                    Next
                    For c = 1 To 3
                        If (FilterSum <> 0) Then
                            a(c) = Int(Math.Round(sum(c) / FilterSum))
                        Else
                            a(c) = Int(Math.Round(sum(c)))
                        End If
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colfil)
                    Else
                        img.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            weightedFilter = locRPSimgResult

        End Function

        Public Function minmaxMedianFilter(ByVal img As System.Drawing.Bitmap, ByVal filnum As Integer) As RPSimgResult
            ' Apply Minimum/Maximum/Median Filter
            ' - img: Input Image
            ' - filnum: -1=Minimum, 1=Maximum, 0=Median
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim wrk As System.Drawing.Bitmap
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim ext(3) As Integer
            Dim med1 As New List(Of Integer)
            Dim med2 As New List(Of Integer)
            Dim med3 As New List(Of Integer)
            Dim u, v As Integer
            Dim i, j As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                minmaxMedianFilter = locRPSimgResult
                Exit Function

            End If

            ' Work Copy
            wrk = img.Clone

            ' Apply Minimum/Maximum Filter
            For v = 1 To h - 2
                For u = 1 To w - 2
                    ' Init
                    col = wrk.GetPixel(u, v)
                    If (filnum = -1 Or filnum = 1) Then
                        ' Min/Max
                        ext(1) = Int(col.R)
                        ext(2) = Int(col.G)
                        ext(3) = Int(col.B)
                    Else
                        ' Median
                        med1.Clear()
                        med2.Clear()
                        med3.Clear()
                    End If
                    For i = -1 To 1
                        For j = -1 To 1
                            col = wrk.GetPixel(u + i, v + j)
                            If (filnum = -1) Then
                                ' Minimum Filter
                                If (Int(col.R) < ext(1)) Then ext(1) = Int(col.R)
                                If (Int(col.G) < ext(2)) Then ext(2) = Int(col.G)
                                If (Int(col.B) < ext(3)) Then ext(3) = Int(col.B)
                            ElseIf (filnum = 1) Then
                                ' Maximum Filter
                                If (Int(col.R) > ext(1)) Then ext(1) = Int(col.R)
                                If (Int(col.G) > ext(2)) Then ext(2) = Int(col.G)
                                If (Int(col.B) > ext(3)) Then ext(3) = Int(col.B)
                            Else
                                ' Median Filter
                                med1.Add(Int(col.R))
                                med2.Add(Int(col.G))
                                med3.Add(Int(col.B))
                            End If
                        Next
                    Next
                    If (filnum = -1 Or filnum = 1) Then
                        ' Min/Max
                        For c = 1 To 3
                            a(c) = ext(c)
                        Next
                    Else
                        ' Median
                        med1.Sort()
                        a(1) = med1.Item(4)
                        med1.Sort()
                        a(2) = med2.Item(4)
                        med1.Sort()
                        a(3) = med3.Item(4)
                    End If
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colfil)
                    Else
                        img.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            minmaxMedianFilter = locRPSimgResult

        End Function

        Public Function unsharpMasking(ByVal imgOrig As System.Drawing.Bitmap, ByVal imgFilter As System.Drawing.Bitmap, ByVal sfac As Double) As RPSimgResult
            ' Unsharp Masking from original and filtered Image
            ' - imgOrig  : original Image
            ' - imgFilter: filtered Image
            ' - sfac: Sharpening Factor (0.25 ... 4.0)
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim sum(3) As Integer
            Dim u, v As Integer
            Dim c As Integer

            ' Bild-Info
            prop = imgFilter.PropertyItems
            w = imgFilter.Width
            h = imgFilter.Height

            ' Check
            If (imgFilter.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = imgFilter.Clone
                unsharpMasking = locRPSimgResult
                Exit Function

            End If

            ' Check
            If (sfac < 0.0) Then sfac = 0.0

            ' Unsharp Masking
            For v = 1 To h - 2
                For u = 1 To w - 2
                    col = imgOrig.GetPixel(u, v)
                    colfil = imgFilter.GetPixel(u, v)
                    a(1) = Int(col.R) + sfac * (Int(col.R) - Int(colfil.R))
                    a(2) = Int(col.G) + sfac * (Int(col.G) - Int(colfil.G))
                    a(3) = Int(col.B) + sfac * (Int(col.B) - Int(colfil.B))
                    For c = 1 To 3
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (imgFilter.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(imgFilter, u, v, colfil)
                    Else
                        imgFilter.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = imgFilter.Clone
            unsharpMasking = locRPSimgResult

        End Function

        Public Function edgeSharpening(ByVal imgOrig As System.Drawing.Bitmap, ByVal imgFilter As System.Drawing.Bitmap, ByVal sfac As Double) As RPSimgResult
            ' Edge Sharpening from original and filtered Image
            ' - imgOrig  : original Image
            ' - imgFilter: filtered Image
            ' - sfac: Sharpening Factor (0.25 ... 4.0)
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim sum(3) As Integer
            Dim u, v As Integer
            Dim c As Integer

            ' Bild-Info
            prop = imgFilter.PropertyItems
            w = imgFilter.Width
            h = imgFilter.Height

            ' Check
            If (imgFilter.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = imgFilter.Clone
                edgeSharpening = locRPSimgResult
                Exit Function

            End If

            ' Check
            If (sfac < 0.0) Then sfac = 0.0

            ' Unsharp Masking
            For v = 1 To h - 2
                For u = 1 To w - 2
                    col = imgOrig.GetPixel(u, v)
                    colfil = imgFilter.GetPixel(u, v)
                    a(1) = Int(col.R) - sfac * Int(colfil.R)
                    a(2) = Int(col.G) - sfac * Int(colfil.G)
                    a(3) = Int(col.B) - sfac * Int(colfil.B)
                    For c = 1 To 3
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (imgFilter.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(imgFilter, u, v, colfil)
                    Else
                        imgFilter.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = imgFilter.Clone
            edgeSharpening = locRPSimgResult

        End Function

        Public Function gradientFilter(ByVal img As System.Drawing.Bitmap, ByVal filnum As Integer, ByVal strengthOrOrient As Integer) As RPSimgResult
            ' Apply Gradient Filter
            ' - img: Input Image
            ' - filnum: 31=Sobel Filter 3x3, 32=Prewitt Filter 3x3
            ' - weightOrOrient: 0=Edge Strength, 1=Edge Orientation
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim wrk As System.Drawing.Bitmap
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim o(3) As Integer
            Dim sum(6) As Integer
            Dim u, v As Integer
            Dim i, j As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                gradientFilter = locRPSimgResult
                Exit Function

            End If

            ' Work Copy
            wrk = img.Clone

            ' Create Filter
            Dim Filterx(,) As Double
            Dim FilterSum As Integer
            Dim FilterSize As Integer
            If (filnum = 31) Then
                ' Sobel Filter 3x3
                FilterSize = 2
                FilterSum = 0
                Dim Filtery(,) As Double = {{-1, 0, 1}, {-2, 0, 2}, {-1, 0, 1}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 32) Then
                ' Prewitt Filter 3x3
                FilterSize = 2
                FilterSum = 0
                Dim Filtery(,) As Double = {{-1, 0, 1}, {-1, 0, 1}, {-1, 0, 1}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            Else
                ' Sobel Filter
                FilterSize = 2
                FilterSum = 0
                Dim Filtery(,) As Double = {{-1, 0, 1}, {-2, 0, 2}, {-1, 0, 1}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            End If

            ' Apply gradient Filter
            For v = FilterSize / 2 To h - 1 - FilterSize / 2
                For u = FilterSize / 2 To w - 1 - FilterSize / 2
                    sum(1) = 0
                    sum(2) = 0
                    sum(3) = 0
                    sum(4) = 0
                    sum(5) = 0
                    sum(6) = 0
                    For i = -FilterSize / 2 To FilterSize / 2
                        For j = -FilterSize / 2 To FilterSize / 2
                            col = wrk.GetPixel(u + i, v + j)
                            sum(1) = sum(1) + Int(col.R * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                            sum(2) = sum(2) + Int(col.G * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                            sum(3) = sum(3) + Int(col.B * Filterx(j + FilterSize / 2, i + FilterSize / 2))
                            sum(4) = sum(4) + Int(col.R * Filterx(i + FilterSize / 2, j + FilterSize / 2))
                            sum(5) = sum(5) + Int(col.G * Filterx(i + FilterSize / 2, j + FilterSize / 2))
                            sum(6) = sum(6) + Int(col.B * Filterx(i + FilterSize / 2, j + FilterSize / 2))
                        Next
                    Next
                    For c = 1 To 3
                        If (FilterSum <> 0) Then
                            a(c) = Int(Math.Sqrt(sum(c) ^ 2 + sum(c + 3) ^ 2) / FilterSum)  ' Edge Strength
                            o(c) = Math.Atan2(sum(c + 3), sum(c)) / (2 * Math.PI) * 255     ' Edge Orientation
                        Else
                            a(c) = Int(Math.Sqrt(sum(c) ^ 2 + sum(c + 3) ^ 2))              ' Edge Strength
                            o(c) = Math.Atan2(sum(c + 3), sum(c)) / (2 * Math.PI) * 255     ' Edge Orientation
                        End If
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                        If (o(c) > amax) Then o(c) = amax
                        If (o(c) < 0) Then o(c) = 0
                    Next
                    If (strengthOrOrient = 1) Then
                        colfil = Color.FromArgb(o(1), o(2), o(3))
                    Else
                        colfil = Color.FromArgb(a(1), a(2), a(3))
                    End If
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colfil)
                    Else
                        img.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            gradientFilter = locRPSimgResult

        End Function

        Public Function morphologFilter(ByVal img As System.Drawing.Bitmap, ByVal type As Integer, ByVal filnum As Integer) As RPSimgResult
            ' Apply Morphological Filter
            ' - img: Input Image
            ' - type: -1=Erosion, 1=Dilatation
            ' - filnum: 20=4-Neighbourhood 3x3, 30=8-Neighbourhood 3x3, 50=Disk-Formed Filter 5x5
            Dim locRPSimgResult As New RPSimgResult
            Dim locRPSindexedImg As New RPSindexedImg
            Dim wrk As System.Drawing.Bitmap
            Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
            Dim w, h As Integer     ' image sizes
            Dim col As System.Drawing.Color
            Dim colfil As System.Drawing.Color
            Dim a(3) As Integer
            Dim val(3) As Integer
            Dim ext(3) As Integer
            Dim u, v As Integer
            Dim i, j As Integer
            Dim c As Integer

            ' Bild-Info
            prop = img.PropertyItems
            w = img.Width
            h = img.Height

            ' Check
            If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                MsgBox("Not possible for this Image Format", MsgBoxStyle.Information, "Incompatible Format")
                locRPSimgResult.ans = False
                locRPSimgResult.img = img.Clone
                morphologFilter = locRPSimgResult
                Exit Function

            End If

            ' Work Copy
            wrk = img.Clone

            ' Create Filter
            Dim Filterx(,) As Double
            Dim FilterSum As Integer
            Dim FilterSize As Integer
            If (filnum = 20) Then
                ' 4-Neighbourhood 3x3
                FilterSize = 2
                FilterSum = 5
                Dim Filtery(,) As Double = {{0, 1, 0}, {1, 1, 1}, {0, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 30) Then
                ' 8-Neighbourhood 3x3
                FilterSize = 2
                FilterSum = 9
                Dim Filtery(,) As Double = {{1, 1, 1}, {1, 1, 1}, {1, 1, 1}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            ElseIf (filnum = 50) Then
                ' Disk-Formed 5x5
                FilterSize = 4
                FilterSum = 9
                Dim Filtery(,) As Double = {{0, 1, 1, 1, 0}, {1, 1, 1, 1, 1}, {1, 1, 1, 1, 1}, {1, 1, 1, 1, 1}, {0, 1, 1, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            Else
                ' 4-Neighbourhood 3x3
                FilterSize = 2
                FilterSum = 5
                Dim Filtery(,) As Double = {{0, 1, 0}, {1, 1, 1}, {0, 1, 0}}
                ReDim Filterx(FilterSize, FilterSize)
                Filterx = Filtery
            End If

            ' Apply Morphological Filter
            For v = FilterSize / 2 To h - 1 - FilterSize / 2
                For u = FilterSize / 2 To w - 1 - FilterSize / 2
                    If (type < 0) Then
                        ' Erosion
                        ext(1) = 255
                        ext(2) = 255
                        ext(3) = 255
                    Else
                        ' Dilatation
                        ext(1) = 0
                        ext(2) = 0
                        ext(3) = 0
                    End If
                    For i = -FilterSize / 2 To FilterSize / 2
                        For j = -FilterSize / 2 To FilterSize / 2
                            col = wrk.GetPixel(u + i, v + j)
                            val(1) = Int(col.R) + Math.Sign(type) * Filterx(j + FilterSize / 2, i + FilterSize / 2)
                            val(2) = Int(col.G) + Math.Sign(type) * Filterx(j + FilterSize / 2, i + FilterSize / 2)
                            val(3) = Int(col.B) + Math.Sign(type) * Filterx(j + FilterSize / 2, i + FilterSize / 2)
                            If (type < 0) Then
                                ' Erosion: minimize
                                For c = 1 To 3
                                    If (val(c) < ext(c)) Then ext(c) = val(c)
                                Next
                            Else
                                ' Dilatation: maximize
                                For c = 1 To 3
                                    If (val(c) > ext(c)) Then ext(c) = val(c)
                                Next
                            End If

                        Next
                    Next
                    For c = 1 To 3
                        a(c) = ext(c)
                        If (a(c) > amax) Then a(c) = amax
                        If (a(c) < 0) Then a(c) = 0
                    Next
                    colfil = Color.FromArgb(a(1), a(2), a(3))
                    If (img.PixelFormat.ToString.EndsWith("Indexed") = True) Then
                        locRPSindexedImg.SetIndexedPixel(img, u, v, colfil)
                    Else
                        img.SetPixel(u, v, colfil)
                    End If
                Next
            Next

            ' Rueckgabe
            locRPSimgResult.ans = True
            locRPSimgResult.img = img.Clone
            morphologFilter = locRPSimgResult

        End Function

    End Class


    Public Class RPSgeoInterp
        ' Calculate Interpolation for geometric Transformations

        Public Class NearestNeighbourInterpolator : Inherits PixelInterpolator
            ' Nearest Neighbour
            Dim img As System.Drawing.Bitmap

            Public Overrides Sub setImageProcessor(ByVal img As Bitmap)
                ' Image
                Me.img = img

            End Sub

            Public Overrides Function getInterpolatedPixel(ByVal pnt As Point2Ddbl) As System.Drawing.Color
                ' retrieve Pixel Value
                Dim u As Integer = Math.Round(pnt.getX())
                Dim v As Integer = Math.Round(pnt.getY())
                If (u >= 0 And v >= 0 And u < img.Width And v < img.Height) Then
                    Return img.GetPixel(u, v)
                Else
                    Return System.Drawing.Color.Black
                End If

            End Function

        End Class


        Public Class BilinearInterpolator : Inherits PixelInterpolator
            ' Bilinear Interpolation
            Dim img As System.Drawing.Bitmap

            Public Overrides Sub setImageProcessor(ByVal img As Bitmap)
                ' Image
                Me.img = img

            End Sub

            Public Overrides Function getInterpolatedPixel(ByVal pnt As Point2Ddbl) As System.Drawing.Color
                ' retrieve Pixel Value
                Dim colint As System.Drawing.Color
                Dim x As Double = pnt.getX()
                Dim y As Double = pnt.getY()
                Dim u As Integer = Math.Floor(x)
                Dim v As Integer = Math.Floor(y)
                Dim aa As Double = x - u
                Dim bb As Double = y - v

                Dim Ap As System.Drawing.Color = System.Drawing.Color.Black
                If (u >= 0 And v >= 0 And u < img.Width And v < img.Height) Then
                    Ap = img.GetPixel(u, v)
                End If
                Dim bp As System.Drawing.Color = System.Drawing.Color.Black
                If ((u + 1) >= 0 And v >= 0 And (u + 1) < img.Width And v < img.Height) Then
                    bp = img.GetPixel(u + 1, v)
                End If
                Dim cp As System.Drawing.Color = System.Drawing.Color.Black
                If (u >= 0 And (v + 1) >= 0 And u < img.Width And (v + 1) < img.Height) Then
                    cp = img.GetPixel(u, v + 1)
                End If
                Dim dp As System.Drawing.Color = System.Drawing.Color.Black
                If ((u + 1) >= 0 And (v + 1) >= 0 And (u + 1) < img.Width And (v + 1) < img.Height) Then
                    dp = img.GetPixel(u + 1, v + 1)
                End If
                Dim A() As Integer = {Ap.R, Ap.G, Ap.B}
                Dim B() As Integer = {Bp.R, Bp.G, Bp.B}
                Dim C() As Integer = {Cp.R, Cp.G, Cp.B}
                Dim D() As Integer = {Dp.R, Dp.G, Dp.B}

                Dim ci As Integer
                Dim G(2) As Integer
                For ci = 0 To 2
                    Dim E As Double = A(ci) + aa * (B(ci) - A(ci))
                    Dim F As Double = C(ci) + aa * (D(ci) - C(ci))
                    G(ci) = Int(E + bb * (F - E))
                Next
                colint = Color.FromArgb(G(0), G(1), G(2))

                Return colint

            End Function

        End Class


        Public Class BicubicInterpolator : Inherits PixelInterpolator
            ' Bicubic Interpolation
            Dim img As System.Drawing.Bitmap
            Dim a As Double = 1

            Public Sub New(ByVal a As Double)
                ' Init
                Me.a = a

            End Sub

            Public Overrides Sub setImageProcessor(ByVal img As Bitmap)
                ' Image
                Me.img = img

            End Sub

            Public Overrides Function getInterpolatedPixel(ByVal pnt As Point2Ddbl) As System.Drawing.Color
                ' retrieve Pixel Value
                Dim colint As System.Drawing.Color
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim u0 As Integer = Math.Floor(x0)  ' use floor to handle negative coordinates too
                Dim v0 As Integer = Math.Floor(y0)

                Dim Aq As System.Drawing.Color
                Dim q() As Double = {0.0, 0.0, 0.0}
                Dim i, j As Integer
                Dim u, v As Integer
                Dim p(2) As Double
                Dim G(2) As Integer
                For j = 0 To 3
                    v = v0 - 1 + j
                    p(0) = 0
                    p(1) = 0
                    p(2) = 0
                    For i = 0 To 3
                        u = u0 - 1 + i
                        Aq = System.Drawing.Color.Black
                        If (u >= 0 And v >= 0 And u < img.Width And v < img.Height) Then
                            Aq = img.GetPixel(u, v)
                        End If
                        p(0) = p(0) + Aq.R * cubic(x0 - u)
                        p(1) = p(1) + Aq.G * cubic(x0 - u)
                        p(2) = p(2) + Aq.B * cubic(x0 - u)
                    Next
                    q(0) = q(0) + p(0) * cubic(y0 - v)
                    q(1) = q(1) + p(1) * cubic(y0 - v)
                    q(2) = q(2) + p(2) * cubic(y0 - v)
                Next
                G(0) = 0
                G(1) = 0
                G(2) = 0
                If (q(0) >= 0 And q(1) >= 0 And q(2) >= 0 And q(0) <= 255 And q(1) <= 255 And q(2) <= 255) Then
                    G(0) = Int(q(0))
                    G(1) = Int(q(1))
                    G(2) = Int(q(2))
                End If
                colint = Color.FromArgb(G(0), G(1), G(2))

                Return colint

            End Function

            Public Function cubic(ByVal x As Double) As Double
                If (x < 0) Then x = -x
                Dim z As Double = 0
                If (x < 1) Then
                    z = (-a + 2) * x * x * x + (a - 3) * x * x + 1
                ElseIf (x < 2) Then
                    z = -a * x * x * x + 5 * a * x * x - 8 * a * x + 4 * a
                End If
                Return z

            End Function

        End Class


        Public Class BitchellNetravaliInterpolator : Inherits PixelInterpolator
            ' Mitchell-Netravali Approximation
            Dim img As System.Drawing.Bitmap

            Public Overrides Sub setImageProcessor(ByVal img As Bitmap)
                ' Image
                Me.img = img

            End Sub

            Public Overrides Function getInterpolatedPixel(ByVal pnt As Point2Ddbl) As System.Drawing.Color
                ' retrieve Pixel Value
                Dim colint As System.Drawing.Color
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim u0 As Integer = Math.Floor(x0)  ' use floor to handle negative coordinates too
                Dim v0 As Integer = Math.Floor(y0)

                Dim Aq As System.Drawing.Color
                Dim q() As Double = {0.0, 0.0, 0.0}
                Dim i, j As Integer
                Dim u, v As Integer
                Dim p(2) As Double
                Dim G(2) As Integer
                For j = 0 To 3
                    v = v0 - 1 + j
                    p(0) = 0
                    p(1) = 0
                    p(2) = 0
                    For i = 0 To 3
                        u = u0 - 1 + i
                        Aq = System.Drawing.Color.Black
                        If (u >= 0 And v >= 0 And u < img.Width And v < img.Height) Then
                            Aq = img.GetPixel(u, v)
                        End If
                        p(0) = p(0) + Aq.R * mitchellNetravali(x0 - u)
                        p(1) = p(1) + Aq.G * mitchellNetravali(x0 - u)
                        p(2) = p(2) + Aq.B * mitchellNetravali(x0 - u)
                    Next
                    q(0) = q(0) + p(0) * mitchellNetravali(y0 - v)
                    q(1) = q(1) + p(1) * mitchellNetravali(y0 - v)
                    q(2) = q(2) + p(2) * mitchellNetravali(y0 - v)
                Next
                G(0) = 0
                G(1) = 0
                G(2) = 0
                If (q(0) >= 0 And q(1) >= 0 And q(2) >= 0 And q(0) <= 255 And q(1) <= 255 And q(2) <= 255) Then
                    G(0) = Int(q(0))
                    G(1) = Int(q(1))
                    G(2) = Int(q(2))
                End If
                colint = Color.FromArgb(G(0), G(1), G(2))

                Return colint

            End Function

            Public Function mitchellNetravali(ByVal x As Double) As Double
                If (x < 0) Then x = -x
                Dim z As Double = 0
                If (x < 1) Then
                    z = 21 * x * x * x - 36 * x * x + 16
                ElseIf (x < 2) Then
                    z = -7 * x * x * x + 36 * x * x - 60 * x + 32
                End If
                Return (z / 18)

            End Function

        End Class



        Public Class PixelInterpolator
            ' Interpolator Base Class
            Dim img As System.Drawing.Bitmap

            Public Overridable Sub setImageProcessor(ByVal img As Bitmap)
                ' Image
                Me.img = img

            End Sub

            Public Overridable Function getInterpolatedPixel(ByVal pnt As Point2Ddbl) As System.Drawing.Color
                ' retrieve Pixel Value

            End Function

        End Class

    End Class


    Public Class RPSgeoMapping
        ' Calculate Mappings for geometric Transformations

        Public Class Translation : Inherits AffineMapping
            ' Translation

            Public Sub New(ByVal dx As Double, ByVal dy As Double)
                ' Init
                MyBase.new(1, 0, dx, _
                            0, 1, dy, _
                            False)
            End Sub

        End Class

        Public Class Scaling : Inherits AffineMapping
            ' Scaling

            Public Sub New(ByVal sx As Double, ByVal sy As Double)
                ' Init
                MyBase.new(sx, 0, 0, _
                            0, sy, 0, _
                            False)
            End Sub

        End Class

        Public Class Rotation : Inherits AffineMapping
            ' Rotation

            Public Sub New(ByVal alpha As Double)
                ' Init
                MyBase.new(Math.Cos(alpha), -Math.Sin(alpha), 0, _
                            Math.Sin(alpha), Math.Cos(alpha), 0, _
                            False)
            End Sub

        End Class

        Public Class Shear : Inherits AffineMapping
            ' Shear

            Public Sub New(ByVal bx As Double, ByVal by As Double)
                ' Init
                MyBase.new(1, bx, 0, _
                            by, 1, 0, _
                            False)
            End Sub

        End Class

        Public Class AffineMapping : Inherits LinearMapping
            ' Affine Mapping

            Public Sub New(ByVal a11 As Double, ByVal a12 As Double, ByVal a13 As Double, _
                            ByVal a21 As Double, ByVal a22 As Double, ByVal a23 As Double, _
                            ByVal inv As Boolean)
                ' Init
                MyBase.new(a11, a12, a13, a21, a22, a23, 0, 0, 1, inv)

            End Sub

            Public Function makeMapping(ByVal A1 As Point2Ddbl, ByVal A2 As Point2Ddbl, ByVal A3 As Point2Ddbl, _
                                        ByVal B1 As Point2Ddbl, ByVal B2 As Point2Ddbl, ByVal B3 As Point2Ddbl) As AffineMapping
                ' Static -> a11, a12, a21, a22, a13, a23 public
                ' Calculate Coefficients
                Dim ax1, ax2, ax3 As Double
                Dim ay1, ay2, ay3 As Double
                Dim bx1, bx2, bx3 As Double
                Dim by1, by2, by3 As Double

                ax1 = A1.getX() : ax2 = A2.getX() : ax3 = A3.getX()
                ay1 = A1.getY() : ay2 = A2.getY() : ay3 = A3.getY()
                bx1 = B1.getX() : bx2 = B2.getX() : bx3 = B3.getX()
                by1 = B1.getY() : by2 = B2.getY() : by3 = B3.getY()

                Dim S As Double = ax1 * (ay3 - ay2) + ax2 * (ay1 - ay3) + ax3 * (ay2 - ay1)
                Dim a11 As Double = (ay1 * (bx2 - bx3) + ay2 * (bx3 - bx1) + ay3 * (bx1 - bx2)) / S
                Dim a12 As Double = (ax1 * (bx3 - bx2) + ax2 * (bx1 - bx3) + ax3 * (bx2 - bx1)) / S
                Dim a21 As Double = (ay1 * (by2 - by3) + ay2 * (by3 - by1) + ay3 * (by1 - by2)) / S
                Dim a22 As Double = (ax1 * (by3 - by2) + ax2 * (by1 - by3) + ax3 * (by2 - by1)) / S
                Dim a13 As Double = (ax1 * (ay3 * bx2 - ay2 * bx3) + ax2 * (ay1 * bx3 - ay3 * bx1) + ax3 * (ay2 * bx1 - ay1 * bx2)) / S
                Dim a23 As Double = (ax1 * (ay3 * by2 - ay2 * by3) + ax2 * (ay1 * by3 - ay3 * by1) + ax3 * (ay2 * by1 - ay1 * by2)) / S

                MyBase.Update(a11, a12, a13, a21, a22, a23, 0, 0, 0)
                Return New AffineMapping(a11, a12, a13, a21, a22, a23, False)

            End Function

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Affine Mapping for Bitmap Image
                Dim x, y As Double

                x = pnt.getX : y = pnt.getY

                Dim xx As Double = a11 * x + a12 * y + a13
                Dim yy As Double = a21 * x + a22 * y + a23

                Dim pntnew As New Point2Ddbl(0, 0)
                pntnew.setLocation(xx, yy)
                Return pntnew

            End Function

            Public Overloads Function getInverse() As AffineMapping
                ' Inverse

                Dim r As Double = a11 * a22 - a12 * a21
                Dim b11 As Double = a22 / r
                Dim b12 As Double = -a12 / r
                Dim b13 As Double = (a12 * a23 - a13 * a22) / r
                Dim b21 As Double = -a21 / r
                Dim b22 As Double = a11 / r
                Dim b23 As Double = (a13 * a21 - a11 * a23) / r
                Return New AffineMapping(b11, b12, b13, b21, b22, b23, True)

            End Function

        End Class

        Public Class ProjectiveMapping : Inherits LinearMapping
            ' Projective Mapping

            Public Sub New(ByVal a11 As Double, ByVal a12 As Double, ByVal a13 As Double, _
                            ByVal a21 As Double, ByVal a22 As Double, ByVal a23 As Double, _
                            ByVal a31 As Double, ByVal a32 As Double, ByVal inv As Boolean)
                ' Init
                MyBase.new(a11, a12, a13, a21, a22, a23, a31, a32, 1, inv)

            End Sub

            Public Function makeMapping(ByVal A1 As Point2Ddbl, ByVal A2 As Point2Ddbl, _
                                        ByVal A3 As Point2Ddbl, ByVal A4 As Point2Ddbl) As ProjectiveMapping
                ' creates the projective mapping from the unit square S to
                ' the arbitrary quadrilateral Q given by points P1 ... P4:
                ' Static -> a11, a12, a21, a22, a13, a23, a31, a32 public
                ' Calculate Coefficients
                Dim x1, x2, x3, x4 As Double
                Dim y1, y2, y3, y4 As Double

                x1 = A1.getX() : x2 = A2.getX() : x3 = A3.getX() : x4 = A4.getX()
                y1 = A1.getY() : y2 = A2.getY() : y3 = A3.getY() : y4 = A4.getY()
                Dim S As Double = (x2 - x3) * (y4 - y3) - (x4 - x3) * (y2 - y3)
                Dim a31 As Double = ((x1 - x2 + x3 - x4) * (y4 - y3) - (y1 - y2 + y3 - y4) * (x4 - x3)) / S
                Dim a32 As Double = ((y1 - y2 + y3 - y4) * (x2 - x3) - (x1 - x2 + x3 - x4) * (y2 - y3)) / S
                Dim a11 As Double = x2 - x1 + a31 * x2
                Dim a12 As Double = x4 - x1 + a32 * x4
                Dim a13 As Double = x1
                Dim a21 As Double = y2 - y1 + a31 * y2
                Dim a22 As Double = y4 - y1 + a32 * y4
                Dim a23 As Double = y1

                MyBase.Update(a11, a12, a13, a21, a22, a23, a31, a32, 0)
                Return New ProjectiveMapping(a11, a12, a13, a21, a22, a23, a31, a32, False)

            End Function

            Public Function makeMapping(ByVal A1 As Point2Ddbl, ByVal A2 As Point2Ddbl, ByVal A3 As Point2Ddbl, ByVal A4 As Point2Ddbl, _
                                        ByVal B1 As Point2Ddbl, ByVal B2 As Point2Ddbl, ByVal B3 As Point2Ddbl, ByVal B4 As Point2Ddbl) As ProjectiveMapping
                ' creates the projective mapping between arbitrary quadrilaterals Qa, Qb
                ' via the unit square: Qa -> S -> Qb
                ' Static -> a11, a12, a21, a22, a13, a23, a31, a32 public
                ' Calculate Coefficients

                Dim T1 As ProjectiveMapping = makeMapping(A1, A2, A3, A4)
                Dim T2 As ProjectiveMapping = makeMapping(B1, B2, B3, B4)
                Dim T1i As LinearMapping = DirectCast(T1, LinearMapping).getInverse
                Dim T As LinearMapping = T1i.concat(T2)
                T.isInverse = False

                MyBase.Update(T.a11, T.a12, T.a13, T.a21, T.a22, T.a23, T.a31, T.a32, T.a33)
                Return (DirectCast(T, ProjectiveMapping))

            End Function

        End Class

        Public Class LinearMapping : Inherits Mapping
            ' Linear Mapping
            Public a11, a12, a13 As Double
            Public a21, a22, a23 As Double
            Public a31, a32, a33 As Double

            Public Sub New()
                ' Constructor
                a11 = 1
                a22 = 1
                a33 = 1

            End Sub

            Public Sub New(ByVal a11 As Double, ByVal a12 As Double, ByVal a13 As Double, _
                           ByVal a21 As Double, ByVal a22 As Double, ByVal a23 As Double, _
                           ByVal a31 As Double, ByVal a32 As Double, ByVal a33 As Double, _
                           ByVal inv As Boolean)

                ' Init
                Me.a11 = a11 : Me.a12 = a12 : Me.a13 = a13
                Me.a21 = a21 : Me.a22 = a22 : Me.a23 = a23
                Me.a31 = a31 : Me.a32 = a32 : Me.a33 = a33
                isInverse = inv

            End Sub

            Public Sub Update(ByVal a11 As Double, ByVal a12 As Double, ByVal a13 As Double, _
                           ByVal a21 As Double, ByVal a22 As Double, ByVal a23 As Double, _
                           ByVal a31 As Double, ByVal a32 As Double, ByVal a33 As Double)

                ' Init
                Me.a11 = a11 : Me.a12 = a12 : Me.a13 = a13
                Me.a21 = a21 : Me.a22 = a22 : Me.a23 = a23
                Me.a31 = a31 : Me.a32 = a32 : Me.a33 = a33

            End Sub

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Linear Mapping for Bitmap Image
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim h As Double = (a31 * x0 + a32 * y0 + a33)
                Dim x1 As Double = (a11 * x0 + a12 * y0 + a13) / h
                Dim y1 As Double = (a21 * x0 + a22 * y0 + a23) / h
                pnt.setLocation(x1, y1)
                Return pnt

            End Function

            Public Overloads Function getInverse() As Mapping
                ' Inverse Mapping
                Dim lm As New LinearMapping
                lm = DirectCast(DirectCast(Me, Object), LinearMapping).duplicate  'works for all subclasses too
                Dim det As Double = a11 * a22 * a33 + a12 * a23 * a31 + a13 * a21 * a32 - _
                                    a11 * a23 * a32 - a12 * a21 * a33 - a13 * a22 * a31
                lm.a11 = (a22 * a33 - a23 * a32) / det
                lm.a12 = (a13 * a32 - a12 * a33) / det
                lm.a13 = (a12 * a23 - a13 * a22) / det
                lm.a21 = (a23 * a31 - a21 * a33) / det
                lm.a22 = (a11 * a33 - a13 * a31) / det
                lm.a23 = (a13 * a21 - a11 * a23) / det
                lm.a31 = (a21 * a32 - a22 * a31) / det
                lm.a32 = (a12 * a31 - a11 * a32) / det
                lm.a33 = (a11 * a22 - a12 * a21) / det
                lm.isInverse = Not isInverse
                Return lm

            End Function

            Public Function concat(ByVal B As LinearMapping) As LinearMapping
                ' concatenates THIS transform matrix A with B: C = B*A
                Dim lm As LinearMapping = DirectCast(DirectCast(Me, Object), LinearMapping).duplicate
                lm.a11 = B.a11 * a11 + B.a12 * a21 + B.a13 * a31
                lm.a12 = B.a11 * a12 + B.a12 * a22 + B.a13 * a32
                lm.a13 = B.a11 * a13 + B.a12 * a23 + B.a13 * a33

                lm.a21 = B.a21 * a11 + B.a22 * a21 + B.a23 * a31
                lm.a22 = B.a21 * a12 + B.a22 * a22 + B.a23 * a32
                lm.a23 = B.a21 * a13 + B.a22 * a23 + B.a23 * a33

                lm.a31 = B.a31 * a11 + B.a32 * a21 + B.a33 * a31
                lm.a32 = B.a31 * a12 + B.a32 * a22 + B.a33 * a32
                lm.a33 = B.a31 * a13 + B.a32 * a23 + B.a33 * a33
                Return lm

            End Function

        End Class


        Public Class BilinearMapping : Inherits Mapping
            ' Bilinear Mapping
            Dim a1, a2, a3, a4 As Double
            Dim b1, b2, b3, b4 As Double

            Public Sub New(ByVal a1 As Double, ByVal a2 As Double, ByVal a3 As Double, ByVal a4 As Double, _
                             ByVal b1 As Double, ByVal b2 As Double, ByVal b3 As Double, ByVal b4 As Double, _
                             ByVal inv As Boolean)
                ' Init
                Me.a1 = a1 : Me.a2 = a2 : Me.a3 = a3 : Me.a4 = a4
                Me.b1 = b1 : Me.b2 = b2 : Me.b3 = b3 : Me.b4 = b4
                isInverse = inv

            End Sub

            Public Function makeMapping(ByVal P1 As Point2Ddbl, ByVal P2 As Point2Ddbl, ByVal P3 As Point2Ddbl, ByVal P4 As Point2Ddbl, _
                                        ByVal Q1 As Point2Ddbl, ByVal Q2 As Point2Ddbl, ByVal Q3 As Point2Ddbl, ByVal Q4 As Point2Ddbl) As BilinearMapping
                ' creates the projective mapping from the unit square S to
                ' the arbitrary quadrilateral Q given by points P1 ... P4, Q1 ... Q4:
                ' Static -> a1, a2, a3, a4, b1, b2, b3, b4 public
                ' Calculate Coefficients

                Dim x1, x2, x3, x4 As Double
                Dim y1, y2, y3, y4 As Double

                x1 = P1.getX() : x2 = P2.getX() : x3 = P3.getX() : x4 = P4.getX()
                y1 = Q1.getY() : y2 = Q2.getY() : y3 = Q3.getY() : y4 = Q4.getY()
                Dim a1 = x2 - x1
                Dim a2 = x4 - x1
                Dim a3 = x1 - x2 + x3 - x4
                Dim a4 = x1
                Dim b1 = y2 - y1
                Dim b2 = y4 - y1
                Dim b3 = y1 - y2 + y3 - y4
                Dim b4 = y1

                Return New BilinearMapping(a1, a2, a3, a4, b1, b2, b3, b4, False)

            End Function

            '          Public Function makeInverseMapping(ByVal P1 As Point2Ddbl, ByVal P2 As Point2Ddbl, ByVal P3 As Point2Ddbl, ByVal P4 As Point2Ddbl, _
            '                                              ByVal Q1 As Point2Ddbl, ByVal Q2 As Point2Ddbl, ByVal Q3 As Point2Ddbl, ByVal Q4 As Point2Ddbl) As BilinearMapping
            '              ' map between arbitrary quadrilaterals

            '              'define column vectors X, Y
            'Matrix X = new Matrix(new double[][] {{Q1.getX()},{Q2.getX()},{Q3.getX()},{Q4.getX()}})
            'Matrix Y = new Matrix(new double[][] {{Q1.getY()},{Q2.getY()},{Q3.getY()},{Q4.getY()}})

            '              'define matrix M
            'Matrix M = new Matrix(new double[][]
            '	{{P1.getX(), P1.getY(), P1.getX() * P1.getY(), 1},
            '	 {P2.getX(), P2.getY(), P2.getX() * P2.getY(), 1},
            '	 {P3.getX(), P3.getY(), P3.getX() * P3.getY(), 1},
            '	 {P4.getX(), P4.getY(), P4.getX() * P4.getY(), 1}}
            '	)

            '              Matrix(A = M.solve(X))       ' solve M * A = X (A is unknown)
            '              Matrix(B = M.solve(Y))       ' solve M * B = Y (B is unknown)

            '              Dim a1 As Double = A.get(0, 0) : Dim b1 As Double = B.get(0, 0)
            '              Dim a2 As Double = A.get(1, 0) : Dim b2 As Double = B.get(1, 0)
            '              Dim a3 As Double = A.get(2, 0) : Dim b3 As Double = B.get(2, 0)
            '              Dim a4 As Double = A.get(3, 0) : Dim b4 As Double = B.get(3, 0)

            '              Return New BilinearMapping(a1, a2, a3, a4, b1, b2, b3, b4, True)

            '          End Function

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Bilinear Mapping for Bitmap Image
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim x1 As Double = a1 * x0 + a2 * y0 + a3 * x0 * y0 + a4
                Dim y1 As Double = b1 * x0 + b2 * y0 + b3 * x0 * y0 + b4
                pnt.setLocation(x1, y1)
                Return pnt

            End Function

        End Class


        Public Class TwirlMapping : Inherits Mapping
            ' Twirl Mapping 
            Dim xc, yc, angle, rad As Double

            Public Sub New(ByVal xc As Double, ByVal yc As Double, ByVal angle As Double, ByVal rad As Double, _
                             ByVal inv As Boolean)
                ' Init
                Me.xc = xc : Me.yc = yc
                Me.angle = angle : Me.rad = rad
                isInverse = inv

            End Sub

            Public Function makeInverseMapping(ByVal xc As Double, ByVal yc As Double, ByVal angle As Double, ByVal rad As Double) As TwirlMapping
                ' Inverse Mapping
                Return New TwirlMapping(xc, yc, angle, rad, True)

            End Function

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Twirl Mapping for Bitmap Image
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim dx As Double = x0 - xc
                Dim dy As Double = y0 - yc
                Dim d As Double = Math.Sqrt(dx * dx + dy * dy)
                If (d < rad) Then
                    Dim a As Double = Math.Atan2(dy, dx) + angle * (rad - d) / rad
                    Dim x1 As Double = xc + d * Math.Cos(a)
                    Dim y1 As Double = yc + d * Math.Sin(a)
                    pnt.setLocation(x1, y1)

                End If

                Return pnt

            End Function

        End Class


        Public Class RippleMapping : Inherits Mapping
            ' Ripple Mapping 
            Dim xWavel, yWavel, xAmpl, yAmpl As Double

            Public Sub New(ByVal xWavel As Double, ByVal xAmpl As Double, ByVal yWavel As Double, ByVal yAmpl As Double, _
                             ByVal inv As Boolean)
                ' Init
                Me.xWavel = xWavel / (2 * Math.PI)
                Me.yWavel = yWavel / (2 * Math.PI)
                Me.xAmpl = xAmpl
                Me.yAmpl = yAmpl
                isInverse = inv

            End Sub

            Public Function makeInverseMapping(ByVal xWavel As Double, ByVal xAmpl As Double, ByVal yWavel As Double, ByVal yAmpl As Double) As RippleMapping
                ' Inverse Mapping
                Return New RippleMapping(xWavel, xAmpl, yWavel, yAmpl, True)

            End Function

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Ripple Mapping for Bitmap Image
                Dim x0 As Double = pnt.getX()
                Dim y0 As Double = pnt.getY()
                Dim x1 As Double = x0 + xAmpl * Math.Sin(y0 / xWavel)
                Dim y1 As Double = y0 + yAmpl * Math.Sin(x0 / yWavel)
                pnt.setLocation(x1, y1)

                Return pnt

            End Function

        End Class


        Public Class SphereMapping : Inherits Mapping
            ' Sphere Mapping 
            Dim xc, yc, rad, refIdx As Double

            Public Sub New(ByVal xc As Double, ByVal yc As Double, ByVal rad As Double, ByVal refIdx As Double, _
                             ByVal inv As Boolean)
                ' Init
                Me.xc = xc : Me.yc = yc
                Me.rad = rad : Me.refIdx = refIdx
                isInverse = inv

            End Sub

            Public Function makeInverseMapping(ByVal xc As Double, ByVal yc As Double, ByVal rad As Double, ByVal refIdx As Double) As SphereMapping
                ' Inverse Mapping
                Return New SphereMapping(xc, yc, rad, refIdx, True)

            End Function

            Public Overrides Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
                ' Sphere Mapping for Bitmap Image
                Dim dx As Double = pnt.getX() - xc
                Dim dy As Double = pnt.getY() - yc
                Dim dx2 As Double = dx * dx
                Dim dy2 As Double = dy * dy
                Dim rad2 As Double = rad * rad

                Dim r2 As Double = dx * dx + dy * dy

                If (r2 > 0 And r2 < rad2) Then
                    Dim z2 As Double = rad2 - r2
                    Dim z As Double = Math.Sqrt(z2)

                    Dim xAlpha As Double = Math.Asin(dx / Math.Sqrt(dx2 + z2))
                    Dim xBeta As Double = xAlpha - xAlpha * (1 / refIdx)
                    Dim x1 As Double = pnt.getX() - z * Math.Tan(xBeta)

                    Dim yAlpha As Double = Math.Asin(dy / Math.Sqrt(dy2 + z2))
                    Dim yBeta As Double = yAlpha - yAlpha * (1 / refIdx)
                    Dim y1 As Double = pnt.getY() - z * Math.Tan(yBeta)
                    pnt.setLocation(x1, y1)

                End If

                Return pnt

            End Function

        End Class


        Public MustInherit Class Mapping
            ' Mapping Base Class
            Public isInverse As Boolean = True

            Public Sub New()

            End Sub

            Public MustOverride Function applyTo(ByVal pnt As Point2Ddbl) As Point2Ddbl
            ' abstract

            Public Overridable Function applyTo(ByVal img As Bitmap, ByVal intPol As RPSgeoInterp.PixelInterpolator) As RPSimgResult
                ' Image Mapping for Bitmap Image
                Dim locRPSimgResult As New RPSimgResult
                Dim map As System.Drawing.Bitmap
                Dim prop() As System.Drawing.Imaging.PropertyItem   ' image properties
                Dim w, h As Integer
                Dim col As System.Drawing.Color
                Dim colmap As System.Drawing.Color
                Dim u, v As Integer

                map = img.Clone

                Dim invMap As Mapping = Me.getInverse()    ' get inverse mapping 
                intPol.setImageProcessor(img)

                ' Bild-Info
                prop = img.PropertyItems
                w = img.Width
                h = img.Height

                Dim pt As Point2Ddbl = New Point2Ddbl(0, 0)
                For v = 0 To h - 1
                    For u = 0 To w - 1
                        col = img.GetPixel(u, v)
                        pt.setLocation(u, v)
                        pt = invMap.applyTo(pt)
                        If (Double.IsInfinity(pt.getX) = False And Double.IsInfinity(pt.getY) = False) Then
                            colmap = intPol.getInterpolatedPixel(pt)
                        Else
                            colmap = System.Drawing.Color.Black
                        End If
                        map.SetPixel(u, v, colmap)
                    Next
                Next

                locRPSimgResult.ans = True
                locRPSimgResult.img = map

                Return locRPSimgResult

            End Function

            Public Overridable Function getInverse() As Mapping
                ' Inverse

                If (isInverse = True) Then
                    Return Me
                Else
                    'MsgBox("Cannot invert mapping", MsgBoxStyle.Exclamation, "Error")   ' only linear mappings invert
                    Return Me
                End If

            End Function

            Public Function duplicate() As Mapping
                ' clones any mapping
                Dim newMap As Mapping

                Try
                    newMap = DirectCast(Me.MemberwiseClone(), Mapping)

                Catch ex As Exception
                    MsgBox("Error by cloning", MsgBoxStyle.Exclamation, "Error")

                End Try

                Return newMap

            End Function

        End Class

    End Class


    Public Class Point2Ddbl
        ' Contains a double 2D-Point

        Dim x, y As Double

        Public Sub New(ByVal x As Double, ByVal y As Double)
            ' Constructor
            Me.x = x
            Me.y = y

        End Sub

        Public Sub setLocation(ByVal x As Double, ByVal y As Double)
            ' Init
            Me.x = x
            Me.y = y

        End Sub

        Public Function getX() As Double
            Return Me.x

        End Function

        Public Function getY() As Double
            Return Me.y

        End Function

    End Class



End Module
